import path from "path";
import { fs, getEntryOptions, isRouterV5, isSSGEntry, isUseSSRBundle, logger, SERVER_RENDER_FUNCTION_NAME } from "@modern-js/utils";
import * as templates from "./templates";
import { getClientRoutes, getClientRoutesLegacy } from "./getClientRoutes";
import { FILE_SYSTEM_ROUTES_FILE_NAME, ENTRY_POINT_FILE_NAME, ENTRY_BOOTSTRAP_FILE_NAME } from "./constants";
import { getDefaultImports, getServerLoadersFile, getServerCombinedModueFile } from "./utils";
import { walk } from "./nestedRoutes";
const createImportSpecifier = (specifiers) => {
  let defaults = "";
  const named = [];
  for (const { local, imported } of specifiers) {
    if (local && imported) {
      named.push(`${imported} as ${local}`);
    } else if (local) {
      defaults = local;
    } else {
      named.push(imported);
    }
  }
  if (defaults && named.length) {
    return `${defaults}, { ${named.join(", ")} }`;
  } else if (defaults) {
    return defaults;
  } else {
    return `{ ${named.join(", ")} }`;
  }
};
const createImportStatements = (statements) => {
  const deDuplicated = [];
  const seen = /* @__PURE__ */ new Map();
  for (const { value, specifiers, initialize } of statements) {
    if (!seen.has(value)) {
      deDuplicated.push({
        value,
        specifiers,
        initialize
      });
      seen.set(value, specifiers);
    } else {
      var _deDuplicated_modifyIndex;
      seen.get(value).push(...specifiers);
      const modifyIndex = deDuplicated.findIndex((v) => v.value === value);
      var _deDuplicated_modifyIndex_initialize;
      const originInitialize = (_deDuplicated_modifyIndex_initialize = (_deDuplicated_modifyIndex = deDuplicated[modifyIndex]) === null || _deDuplicated_modifyIndex === void 0 ? void 0 : _deDuplicated_modifyIndex.initialize) !== null && _deDuplicated_modifyIndex_initialize !== void 0 ? _deDuplicated_modifyIndex_initialize : "";
      deDuplicated[modifyIndex].initialize = originInitialize.concat(`
${initialize || ""}`);
    }
  }
  return deDuplicated.map(({ value, specifiers, initialize }) => `import ${createImportSpecifier(specifiers)} from '${value}';
${initialize || ""}`).join("\n");
};
const generateCode = async (appContext, config, entrypoints, api) => {
  const { internalDirectory, srcDirectory, appDirectory, internalDirAlias, internalSrcAlias, packageName } = appContext;
  const hookRunners = api.useHookRunners();
  const isV5 = isRouterV5(config);
  const getRoutes = isV5 ? getClientRoutesLegacy : getClientRoutes;
  const importsStatemets = /* @__PURE__ */ new Map();
  const oldVersion = typeof (config === null || config === void 0 ? void 0 : config.runtime.router) === "object" ? Boolean((config === null || config === void 0 ? void 0 : config.runtime.router).oldVersion) : false;
  await Promise.all(entrypoints.map(generateEntryCode));
  return {
    importsStatemets
  };
  async function generateEntryCode(entrypoint) {
    const { entryName, isMainEntry, isAutoMount, fileSystemRoutes } = entrypoint;
    if (isAutoMount) {
      if (fileSystemRoutes) {
        var _config_output;
        let initialRoutes = [];
        let nestedRoutes = null;
        if (entrypoint.entry) {
          initialRoutes = getRoutes({
            entrypoint,
            srcDirectory,
            srcAlias: internalSrcAlias,
            internalDirectory,
            internalDirAlias
          });
        }
        if (!isV5 && entrypoint.nestedRoutesEntry) {
          nestedRoutes = await walk(entrypoint.nestedRoutesEntry, entrypoint.nestedRoutesEntry, {
            name: internalSrcAlias,
            basename: srcDirectory
          }, entrypoint.entryName, entrypoint.isMainEntry, oldVersion);
          if (nestedRoutes) {
            if (!Array.isArray(nestedRoutes)) {
              nestedRoutes = [
                nestedRoutes
              ];
            }
            for (const route of nestedRoutes) {
              initialRoutes.unshift(route);
            }
          }
        }
        const { routes } = await hookRunners.modifyFileSystemRoutes({
          entrypoint,
          routes: initialRoutes
        });
        const config2 = api.useResolvedConfigContext();
        const ssr = getEntryOptions(entryName, isMainEntry, config2.server.ssr, config2.server.ssrByEntries, packageName);
        const useSSG = isSSGEntry(config2, entryName, entrypoints);
        let mode;
        if (ssr) {
          mode = typeof ssr === "object" ? ssr.mode || "string" : "string";
        }
        if (mode === "stream") {
          const hasPageRoute = routes.some((route) => "type" in route && route.type === "page");
          if (hasPageRoute) {
            logger.error("Streaming ssr is not supported when pages dir exists");
            process.exit(1);
          }
        }
        const { code } = await hookRunners.beforeGenerateRoutes({
          entrypoint,
          code: await templates.fileSystemRoutes({
            routes,
            ssrMode: useSSG ? "string" : mode,
            nestedRoutesEntry: entrypoint.nestedRoutesEntry,
            entryName: entrypoint.entryName,
            internalDirectory,
            splitRouteChunks: config2 === null || config2 === void 0 ? void 0 : (_config_output = config2.output) === null || _config_output === void 0 ? void 0 : _config_output.splitRouteChunks
          })
        });
        if (entrypoint.nestedRoutesEntry && isUseSSRBundle(config2)) {
          const routesServerFile = getServerLoadersFile(internalDirectory, entryName);
          const code2 = templates.routesForServer({
            routes
          });
          await fs.ensureFile(routesServerFile);
          await fs.writeFile(routesServerFile, code2);
        }
        const serverLoaderCombined = templates.ssrLoaderCombinedModule(entrypoints, entrypoint, config2, appContext);
        if (serverLoaderCombined) {
          const serverLoaderFile = getServerCombinedModueFile(internalDirectory, entryName);
          await fs.outputFile(serverLoaderFile, serverLoaderCombined);
        }
        fs.outputFileSync(path.resolve(internalDirectory, `./${entryName}/${FILE_SYSTEM_ROUTES_FILE_NAME}`), code, "utf8");
      }
      const { imports } = await hookRunners.modifyEntryImports({
        entrypoint,
        imports: getDefaultImports({
          entrypoint,
          srcDirectory,
          appDirectory,
          internalSrcAlias,
          internalDirAlias
        })
      });
      importsStatemets.set(entryName, imports);
      const entryFile = path.resolve(internalDirectory, `./${entryName}/${ENTRY_POINT_FILE_NAME}`);
      entrypoint.internalEntry = entryFile;
    }
  }
};
const generateIndexCode = async ({ appContext, api, entrypoints, config, importsStatemets, bundlerConfigs }) => {
  const hookRunners = api.useHookRunners();
  const { mountId } = config.html;
  const { internalDirectory, packageName } = appContext;
  await Promise.all(entrypoints.map(async (entrypoint) => {
    const { entryName, isMainEntry, isAutoMount, customBootstrap, fileSystemRoutes } = entrypoint;
    if (isAutoMount) {
      const { plugins } = await hookRunners.modifyEntryRuntimePlugins({
        entrypoint,
        plugins: [],
        bundlerConfigs
      });
      const { code: renderFunction } = await hookRunners.modifyEntryRenderFunction({
        entrypoint,
        code: templates.renderFunction({
          plugins,
          customBootstrap,
          fileSystemRoutes
        })
      });
      const { exportStatement } = await hookRunners.modifyEntryExport({
        entrypoint,
        exportStatement: "export default AppWrapper;"
      });
      const imports = importsStatemets.get(entryName);
      const code = templates.index({
        mountId,
        imports: createImportStatements(imports),
        renderFunction,
        exportStatement
      });
      const entryFile = path.resolve(internalDirectory, `./${entryName}/${ENTRY_POINT_FILE_NAME}`);
      if (config.source.enableAsyncEntry) {
        let rawAsyncEntryCode = `import('./${ENTRY_BOOTSTRAP_FILE_NAME}');`;
        const ssr = getEntryOptions(entryName, isMainEntry, config.server.ssr, config.server.ssrByEntries, packageName);
        if (ssr) {
          rawAsyncEntryCode = `
        export const ${SERVER_RENDER_FUNCTION_NAME} = async (...args) => {
          let entry = await ${rawAsyncEntryCode};
          if (entry.default instanceof Promise){
            entry = await entry.default;
            return entry.default.${SERVER_RENDER_FUNCTION_NAME}.apply(null, args);
          }
          return entry.${SERVER_RENDER_FUNCTION_NAME}.apply(null, args);
        };
        if(typeof window!=='undefined'){
          ${rawAsyncEntryCode}
        }
        `;
        }
        const { code: asyncEntryCode } = await hookRunners.modifyAsyncEntry({
          entrypoint,
          code: rawAsyncEntryCode
        });
        fs.outputFileSync(entryFile, asyncEntryCode, "utf8");
        const bootstrapFile = path.resolve(internalDirectory, `./${entryName}/${ENTRY_BOOTSTRAP_FILE_NAME}`);
        fs.outputFileSync(bootstrapFile, code, "utf8");
      } else {
        fs.outputFileSync(entryFile, code, "utf8");
      }
    }
  }));
};
export {
  createImportStatements,
  generateCode,
  generateIndexCode
};
