import fs from "fs";
import path from "path";
import { isReact18, normalizeToPosixPath, fs as fse, getCommand } from "@modern-js/utils";
import { transform } from "esbuild";
import { parse } from "es-module-lexer";
import { ACTION_EXPORT_NAME, FILE_SYSTEM_ROUTES_FILE_NAME, JS_EXTENSIONS, LOADER_EXPORT_NAME } from "./constants";
const walkDirectory = (dir) => fs.readdirSync(dir).reduce((previous, filename) => {
  const filePath = path.join(dir, filename);
  if (fs.statSync(filePath).isDirectory()) {
    return [
      ...previous,
      ...walkDirectory(filePath)
    ];
  } else {
    return [
      ...previous,
      filePath
    ];
  }
}, []);
const getDefaultImports = ({ entrypoint, srcDirectory, appDirectory, internalSrcAlias, internalDirAlias }) => {
  const { entryName, fileSystemRoutes, customBootstrap, entry } = entrypoint;
  const imports = [
    {
      specifiers: [
        {
          local: "React"
        }
      ],
      value: "react"
    },
    {
      specifiers: [
        {
          local: "ReactDOM"
        }
      ],
      value: isReact18(path.join(appDirectory)) ? "react-dom/client" : "react-dom"
    },
    {
      specifiers: [
        {
          imported: "createApp"
        },
        {
          imported: "bootstrap"
        }
      ],
      value: "@modern-js/runtime"
    },
    customBootstrap && {
      specifiers: [
        {
          local: "customBootstrap"
        }
      ],
      value: normalizeToPosixPath(customBootstrap.replace(srcDirectory, internalSrcAlias))
    }
  ].filter(Boolean);
  if (fileSystemRoutes) {
    const route = {
      specifiers: [
        {
          imported: "routes"
        }
      ],
      value: normalizeToPosixPath(`${internalDirAlias}/${entryName}/${FILE_SYSTEM_ROUTES_FILE_NAME.replace(".js", "")}`)
    };
    if (fileSystemRoutes.globalApp) {
      imports.push({
        specifiers: [
          {
            local: "App"
          }
        ],
        value: normalizeToPosixPath(fileSystemRoutes.globalApp.replace(srcDirectory, internalSrcAlias))
      });
    } else {
      route.initialize = "const App = false;";
    }
    imports.push(route);
  } else {
    imports.push({
      specifiers: [
        {
          local: "App"
        }
      ],
      value: normalizeToPosixPath(entry.replace(srcDirectory, internalSrcAlias))
    });
  }
  return imports;
};
const isPageComponentFile = (filePath) => {
  if (/\.(d|test|spec|e2e)\.(js|jsx|ts|tsx)$/.test(filePath)) {
    return false;
  }
  if ([
    ".js",
    ".jsx",
    ".ts",
    ".tsx"
  ].includes(path.extname(filePath))) {
    return true;
  }
  return false;
};
const replaceWithAlias = (base, filePath, alias) => {
  if (filePath.includes(base)) {
    return normalizeToPosixPath(path.join(alias, path.relative(base, filePath)));
  } else {
    return filePath;
  }
};
const parseModule = async ({ source, filename }) => {
  let content = source;
  if (JS_EXTENSIONS.some((ext) => filename.endsWith(ext))) {
    const result = await transform(content, {
      loader: path.extname(filename).slice(1),
      format: "esm"
    });
    content = result.code;
  }
  return await parse(content);
};
const hasLoader = async (filename, source) => {
  let content = source;
  if (!source) {
    content = (await fse.readFile(filename, "utf-8")).toString();
  }
  if (content) {
    const [, moduleExports] = await parseModule({
      source: content.toString(),
      filename
    });
    return moduleExports.some((e) => e.n === LOADER_EXPORT_NAME);
  }
  return false;
};
const hasAction = async (filename, source) => {
  let content = source;
  if (!source) {
    content = (await fse.readFile(filename, "utf-8")).toString();
  }
  if (content) {
    const [, moduleExports] = await parseModule({
      source: content.toString(),
      filename
    });
    return moduleExports.some((e) => e.n === ACTION_EXPORT_NAME);
  }
  return false;
};
const getServerLoadersFile = (internalDirectory, entryName) => {
  return path.join(internalDirectory, entryName, "route-server-loaders.js");
};
const getServerCombinedModueFile = (internalDirectory, entryName) => {
  return path.join(internalDirectory, entryName, "server-loader-combined.js");
};
const checkIsBuildCommands = () => {
  const buildCommands = [
    "dev",
    "start",
    "build",
    "inspect",
    "deploy",
    "dev-worker"
  ];
  const command = getCommand();
  return buildCommands.includes(command);
};
const isSubDirOrEqual = (parent, child) => {
  if (parent === child) {
    return true;
  }
  const relative = path.relative(parent, child);
  const isSubdir = relative && !relative.startsWith("..") && !path.isAbsolute(relative);
  return Boolean(isSubdir);
};
export {
  checkIsBuildCommands,
  getDefaultImports,
  getServerCombinedModueFile,
  getServerLoadersFile,
  hasAction,
  hasLoader,
  isPageComponentFile,
  isSubDirOrEqual,
  parseModule,
  replaceWithAlias,
  walkDirectory
};
