import path from "path";
import { lintPlugin } from "@modern-js/plugin-lint";
import { cleanRequireCache, emptyDir, getCommand, getArgv, fs, NESTED_ROUTE_SPEC_FILE, newAction, upgradeAction } from "@modern-js/utils";
import { castArray } from "@modern-js/utils/lodash";
import { getLocaleLanguage } from "@modern-js/plugin-i18n/language-detector";
import analyzePlugin from "./analyze";
import initializePlugin from "./initialize";
import { hooks } from "./hooks";
import { i18n, localeKeys } from "./locale";
import serverBuildPlugin from "./plugins/serverBuild";
import deployPlugin from "./plugins/deploy";
import { restart } from "./utils/restart";
import { generateWatchFiles } from "./utils/generateWatchFiles";
import { mergeConfig } from "@modern-js/core";
import { dev } from "./commands";
export * from "./defineConfig";
export * from "./types";
const devCommand = async (program, api) => {
  const runner = api.useHookRunners();
  const devToolMetas = await runner.registerDev();
  const devProgram = program.command("dev").alias("start").usage("[options]").description(i18n.t(localeKeys.command.dev.describe)).option("-c --config <config>", i18n.t(localeKeys.command.shared.config)).option("-e --entry [entry...]", i18n.t(localeKeys.command.dev.entry)).option("--analyze", i18n.t(localeKeys.command.shared.analyze)).option("--api-only", i18n.t(localeKeys.command.dev.apiOnly)).option("--web-only", i18n.t(localeKeys.command.dev.webOnly)).action(async (options) => {
    const { dev: dev2 } = await import("./commands/dev");
    await dev2(api, options);
  });
  for (const meta of devToolMetas) {
    if (!meta.subCommands) {
      continue;
    }
    for (const subCmd of meta.subCommands) {
      devProgram.command(subCmd).action(async (options = {}) => {
        const { appDirectory } = api.useAppContext();
        const { isTypescript } = await import("@modern-js/utils");
        await runner.beforeDevTask(meta);
        await meta.action(options, {
          isTsProject: isTypescript(appDirectory)
        });
      });
    }
  }
};
const buildCommand = async (program, api) => {
  const runner = api.useHookRunners();
  const platformBuilders = await runner.registerBuildPlatform();
  const buildProgram = program.command("build").usage("[options]").description(i18n.t(localeKeys.command.build.describe)).option("-c --config <config>", i18n.t(localeKeys.command.shared.config)).option("--analyze", i18n.t(localeKeys.command.shared.analyze)).action(async (options) => {
    const { build } = await import("./commands/build");
    await build(api, options);
  });
  for (const platformBuilder of platformBuilders) {
    const platforms = castArray(platformBuilder.platform);
    for (const platform of platforms) {
      buildProgram.command(platform).action(async () => {
        const { appDirectory } = api.useAppContext();
        const { isTypescript } = await import("@modern-js/utils");
        await runner.beforeBuildPlatform(platformBuilders);
        await platformBuilder.build(platform, {
          isTsProject: isTypescript(appDirectory)
        });
      });
    }
  }
};
const appTools = (options = {
  bundler: "webpack"
}) => ({
  name: "@modern-js/app-tools",
  post: [
    "@modern-js/plugin-initialize",
    "@modern-js/plugin-analyze",
    "@modern-js/plugin-ssr",
    "@modern-js/plugin-document",
    "@modern-js/plugin-state",
    "@modern-js/plugin-router",
    "@modern-js/plugin-router-v5",
    "@modern-js/plugin-polyfill"
  ],
  registerHook: hooks,
  usePlugins: [
    initializePlugin({
      bundler: (options === null || options === void 0 ? void 0 : options.bundler) === "experimental-rspack" ? "rspack" : "webpack"
    }),
    analyzePlugin({
      bundler: (options === null || options === void 0 ? void 0 : options.bundler) === "experimental-rspack" ? "rspack" : "webpack"
    }),
    serverBuildPlugin(),
    lintPlugin(),
    deployPlugin()
  ],
  setup: (api) => {
    const appContext = api.useAppContext();
    api.setAppContext({
      ...appContext,
      toolsType: "app-tools"
    });
    const nestedRoutes = {};
    const locale = getLocaleLanguage();
    i18n.changeLanguage({
      locale
    });
    return {
      async beforeConfig() {
        var _userConfig_output;
        const userConfig = api.useConfigContext();
        const appContext2 = api.useAppContext();
        if ((_userConfig_output = userConfig.output) === null || _userConfig_output === void 0 ? void 0 : _userConfig_output.tempDir) {
          api.setAppContext({
            ...appContext2,
            internalDirectory: path.resolve(appContext2.appDirectory, userConfig.output.tempDir)
          });
        }
      },
      async commands({ program }) {
        await devCommand(program, api);
        await buildCommand(program, api);
        program.command("serve").usage("[options]").description(i18n.t(localeKeys.command.serve.describe)).option("--api-only", i18n.t(localeKeys.command.dev.apiOnly)).option("-c --config <config>", i18n.t(localeKeys.command.shared.config)).action(async () => {
          const { start } = await import("./commands/serve");
          await start(api);
        });
        program.command("deploy").usage("[options]").option("-c --config <config>", i18n.t(localeKeys.command.shared.config)).option("-s --skip-build", i18n.t(localeKeys.command.shared.skipBuild)).description(i18n.t(localeKeys.command.deploy.describe)).action(async (options2) => {
          if (!options2.skipBuild) {
            const { build } = await import("./commands/build");
            await build(api);
          }
          const { deploy } = await import("./commands/deploy");
          await deploy(api, options2);
          process.exit(0);
        });
        program.command("new").usage("[options]").description(i18n.t(localeKeys.command.new.describe)).option("--config-file <configFile>", i18n.t(localeKeys.command.shared.config)).option("--lang <lang>", i18n.t(localeKeys.command.new.lang)).option("-c, --config <config>", i18n.t(localeKeys.command.new.config)).option("-d, --debug", i18n.t(localeKeys.command.new.debug), false).option("--dist-tag <tag>", i18n.t(localeKeys.command.new.distTag)).option("--registry", i18n.t(localeKeys.command.new.registry)).option("--no-need-install", i18n.t(localeKeys.command.shared.noNeedInstall)).action(async (options2) => {
          await newAction({
            ...options2,
            locale: options2.lang || locale
          }, "mwa");
        });
        program.command("inspect").description("inspect the internal configs").option(`--env <env>`, i18n.t(localeKeys.command.inspect.env), "development").option("--output <output>", i18n.t(localeKeys.command.inspect.output), "/").option("--verbose", i18n.t(localeKeys.command.inspect.verbose)).option("-c --config <config>", i18n.t(localeKeys.command.shared.config)).action(async (options2) => {
          const { inspect } = await import("./commands/inspect");
          inspect(api, options2);
        });
        program.command("upgrade").allowUnknownOption().option("-h --help", "Show help").action(async () => {
          await upgradeAction();
        });
      },
      async prepare() {
        const command = getCommand();
        if (command === "deploy") {
          const isSkipBuild = [
            "-s",
            "--skip-build"
          ].some((tag) => {
            return getArgv().includes(tag);
          });
          if (isSkipBuild) {
            return;
          }
        }
        if (command === "dev" || command === "start" || command === "build" || command === "deploy") {
          const resolvedConfig = api.useResolvedConfigContext();
          if (resolvedConfig.output.cleanDistPath) {
            const appContext2 = api.useAppContext();
            await emptyDir(appContext2.distDirectory);
          }
        }
      },
      async watchFiles() {
        const appContext2 = api.useAppContext();
        const config = api.useResolvedConfigContext();
        return await generateWatchFiles(appContext2, config.source.configDir);
      },
      // 这里会被 core/initWatcher 监听的文件变动触发，如果是 src 目录下的文件变动，则不做 restart
      async fileChange(e) {
        const { filename, eventType, isPrivate } = e;
        if (!isPrivate && (eventType === "change" || eventType === "unlink")) {
          const { closeServer } = await import("./utils/createServer");
          await closeServer();
          await restart(api.useHookRunners(), filename);
        }
      },
      async beforeRestart() {
        cleanRequireCache([
          require.resolve("./analyze")
        ]);
      },
      async modifyFileSystemRoutes({ entrypoint, routes }) {
        nestedRoutes[entrypoint.entryName] = routes;
        return {
          entrypoint,
          routes
        };
      },
      async beforeGenerateRoutes({ entrypoint, code }) {
        const { distDirectory } = api.useAppContext();
        await fs.outputJSON(path.resolve(distDirectory, NESTED_ROUTE_SPEC_FILE), nestedRoutes);
        return {
          entrypoint,
          code
        };
      }
    };
  }
});
var src_default = appTools;
export {
  appTools,
  buildCommand,
  src_default as default,
  dev,
  devCommand,
  mergeConfig
};
