import path, { isAbsolute } from "node:path";
import { nodeFileTrace, resolve } from "@vercel/nft";
import { fs as fse, pkgUp } from "@modern-js/utils";
import { readPackageJSON } from "pkg-types";
import { parseNodeModulePath } from "mlly";
import { linkPackage, writePackage } from "./utils";
const readDirRecursive = async (dir) => {
  const files = await fse.readdir(dir, {
    withFileTypes: true
  });
  const filesAndDirs = await Promise.all(files.map(async (file) => {
    const resPath = path.resolve(dir, file.name);
    return file.isDirectory() ? readDirRecursive(resPath) : resPath;
  }));
  return filesAndDirs.flat();
};
async function isFile(file) {
  try {
    const stat = await fse.stat(file);
    return stat.isFile();
  } catch (error) {
    if (error.code === "ENOENT") {
      return false;
    }
    throw error;
  }
}
const findEntryFiles = async (rootDir) => {
  const files = await readDirRecursive(rootDir);
  return files;
};
const handleDependencies = async (appDir, serverRootDir, include) => {
  const base = "/";
  const entryFiles = await findEntryFiles(serverRootDir);
  const includeEntries = include.map((item) => {
    if (isAbsolute(item)) {
      return item;
    }
    try {
      return require.resolve(item);
    } catch (error) {
    }
    return item;
  });
  const _fileTrace = await nodeFileTrace(entryFiles.concat(includeEntries), {
    base,
    processCwd: serverRootDir,
    resolve: async (id, parent, job, isCjs) => {
      if (id.startsWith("@modern-js/prod-server")) {
        return require.resolve(id, {
          paths: [
            require.resolve("@modern-js/app-tools")
          ]
        });
      } else {
        return resolve(id, parent, job, isCjs);
      }
    }
  });
  const currentProjectModules = path.join(appDir, "node_modules");
  const _resolveTracedPath = (p) => fse.realpath(path.resolve(base, p));
  const tracedFiles = Object.fromEntries(await Promise.all([
    ..._fileTrace.reasons.entries()
  ].map(async ([_path, reasons]) => {
    if (reasons.ignored) {
      return;
    }
    const filePath = await _resolveTracedPath(_path);
    if (filePath.startsWith(serverRootDir) || filePath.startsWith(appDir) && !filePath.startsWith(currentProjectModules)) {
      return;
    }
    if (!await isFile(filePath)) {
      return;
    }
    let baseDir;
    let pkgName;
    let subpath;
    let pkgPath;
    if (filePath.includes("node_modules")) {
      const parsed = parseNodeModulePath(filePath);
      baseDir = parsed.dir;
      pkgName = parsed.name;
      subpath = parsed.subpath;
      pkgPath = path.join(baseDir, pkgName);
    } else {
      const MODERN_UTILS_PATH = "packages/toolkit/utils";
      const MODERN_UTILS_PATH_REGEX = new RegExp(`(.*${MODERN_UTILS_PATH})`);
      const match = filePath.match(MODERN_UTILS_PATH_REGEX);
      const packageJsonPath = match ? path.join(match[0], "package.json") : await pkgUp({
        cwd: path.dirname(filePath)
      });
      if (packageJsonPath) {
        const packageJson = await fse.readJSON(packageJsonPath);
        pkgPath = baseDir = path.dirname(packageJsonPath);
        subpath = path.relative(baseDir, filePath);
        pkgName = packageJson.name;
      }
    }
    if (!baseDir) {
      return;
    }
    const parents = await Promise.all([
      ...reasons.parents
    ].map((p) => _resolveTracedPath(p)));
    const tracedFile = {
      path: filePath,
      parents,
      subpath,
      pkgName,
      pkgPath
    };
    return [
      filePath,
      tracedFile
    ];
  })).then((r) => r.filter(Boolean)));
  const tracedPackages = {};
  for (const tracedFile of Object.values(tracedFiles)) {
    const { pkgName } = tracedFile;
    let tracedPackage = tracedPackages[pkgName];
    let pkgJSON = await readPackageJSON(tracedFile.pkgPath, {
      cache: true
    }).catch(() => {
    });
    if (!pkgJSON) {
      pkgJSON = {
        name: pkgName,
        version: "0.0.0"
      };
    }
    if (!tracedPackage) {
      tracedPackage = {
        name: pkgName,
        versions: {}
      };
      tracedPackages[pkgName] = tracedPackage;
    }
    let tracedPackageVersion = tracedPackage.versions[pkgJSON.version];
    if (!tracedPackageVersion) {
      tracedPackageVersion = {
        path: tracedFile.pkgPath,
        files: [],
        pkgJSON
      };
      tracedPackage.versions[pkgJSON.version] = tracedPackageVersion;
    }
    tracedFile.path.startsWith(tracedFile.pkgPath) && tracedPackageVersion.path === tracedFile.pkgPath && tracedPackageVersion.files.push(tracedFile.path);
    tracedFile.pkgName = pkgName;
    tracedFile.pkgVersion = pkgJSON.version;
  }
  const findPackageParents = (pkg, version) => {
    const versionFiles = pkg.versions[version].files.map((path2) => tracedFiles[path2]);
    const parentPkgs = [
      ...new Set(versionFiles.flatMap((file) => file.parents.map((parentPath) => {
        const parentFile = tracedFiles[parentPath];
        if (parentFile.pkgName === pkg.name) {
          return null;
        }
        return `${parentFile.pkgName}@${parentFile.pkgVersion}`;
      }).filter(Boolean)))
    ];
    return parentPkgs;
  };
  const multiVersionPkgs = {};
  const singleVersionPackages = [];
  for (const tracedPackage of Object.values(tracedPackages)) {
    const versions = Object.keys(tracedPackage.versions);
    if (versions.length === 1) {
      singleVersionPackages.push(tracedPackage.name);
      continue;
    }
    multiVersionPkgs[tracedPackage.name] = {};
    for (const version of versions) {
      multiVersionPkgs[tracedPackage.name][version] = findPackageParents(tracedPackage, version);
    }
  }
  await Promise.all(singleVersionPackages.map((pkgName) => {
    const pkg = tracedPackages[pkgName];
    const version = Object.keys(pkg.versions)[0];
    return writePackage(pkg, version, serverRootDir);
  }));
  console.log("multiVersionPkgs111111111", multiVersionPkgs);
  for (const [pkgName, pkgVersions] of Object.entries(multiVersionPkgs)) {
    const versionEntires = Object.entries(pkgVersions).sort(
      // TODO: remove ts-ignore
      // @ts-expect-error
      ([v1, p1], [v2, p2]) => {
        if (p1.length === 0) {
          return -1;
        }
        if (p2.length === 0) {
          return 1;
        }
      }
    );
    for (const [version, parentPkgs] of versionEntires) {
      const pkg = tracedPackages[pkgName];
      const pkgDestPath = `.modernjs/${pkgName}@${version}/node_modules/${pkgName}`;
      await writePackage(pkg, version, serverRootDir, pkgDestPath);
      await linkPackage(pkgDestPath, `${pkgName}`, serverRootDir);
      for (const parentPkg of parentPkgs) {
        const parentPkgName = parentPkg.replace(/@[^@]+$/, "");
        await (multiVersionPkgs[parentPkgName] ? linkPackage(pkgDestPath, `.modernjs/${parentPkg}/node_modules/${pkgName}@${version}`, path.join(serverRootDir, "node_modules")) : linkPackage(pkgDestPath, `${parentPkgName}/node_modules/${pkgName}@${version}`, path.join(serverRootDir, "node_modules")));
      }
    }
  }
  const projectPkg = await readPackageJSON(serverRootDir).catch(() => ({}));
  const outputPkgPath = path.join(serverRootDir, "package.json");
  await fse.writeJSON(outputPkgPath, {
    name: `${projectPkg.name || "modernjs-project"}-prod`,
    version: projectPkg.version || "0.0.0",
    private: true,
    dependencies: Object.fromEntries([
      ...Object.values(tracedPackages).map((pkg) => [
        pkg.name,
        Object.keys(pkg.versions)[0]
      ])
    ].sort(([a], [b]) => a.localeCompare(b)))
  });
};
export {
  handleDependencies
};
